import json
import logging
import functools
from collections import namedtuple
from typing import List, NamedTuple

from . import client

logger = logging.getLogger(__name__)

_fields = ["name", "status", "mac_address", "description", "link_speed"]
_header = ["Name", "Status", "MAC Address", "Description", "Link Speed"]

# The field names returned by Powershell - these are only
# used when loading the records into namedtuples. Having all
# lowercase attributes on the tuples is easier to use.
_powershell_return_fields = ["Name", "Status", "MacAddress", "InterfaceDescription", "LinkSpeed"]
_network_adapter = namedtuple("network_adapter", _fields)  # type: ignore


def _as_objects(func):
    @functools.wraps(func)
    def _fn(connection, *args, **kwargs) -> List[NamedTuple]:
        output = func(connection, *args, **kwargs)
        if not output:
            return []
        json_data = json.loads(output)
        if not isinstance(json_data, list):
            json_data = [json_data]
        return [
            _network_adapter(*[adapter.get(field, "") for field in _powershell_return_fields])  # type: ignore
            for adapter in json_data
        ]

    return _fn


@_as_objects
def network_adapters(connection):
    """ Get all network adapters.

    :returns: A list of :ref:`adapters:Adapter Records`.

    Example:

    .. code-block:: python

        with session() as connection:
            all_adapters = adapters.network_adapters(connection)

    """
    output = client.run_powershell(
        connection,
        script="Get-NetAdapter | Select {fields} | ConvertTo-Json".format(fields=", ".join(_powershell_return_fields)),
    )
    logger.info(f"got network adapters: {output}")
    return output


def adapter_properties(connection, *, adapter: str):
    # Doesn't use namedtuples as we don't know for sure which
    # properties will be returned for different adapter types.
    return {
        pair["DisplayName"]: pair["DisplayValue"]
        for pair in json.loads(
            client.run_powershell(
                connection,
                script=(
                    f"Get-NetAdapterAdvancedProperty -Name {adapter} | "
                    "Select DisplayName, DisplayValue | ConvertTo-Json"
                ),
            )
        )
    }
