"""
Users

This module contains actions for auditing _local_ users on a Windows Server.
"""
import logging

from typing import List, Dict, Any, Union, Optional

from ppa import ui
from windows_server import client
from playbook_tools import tables

from . import exceptions
from .records import WindowsServer, User, UserData
from ._wrappers import winrm, json_load, error_handler, member_only


logger = logging.getLogger(__name__)

USER_TABLE_HEADER = ["Name", "Full Name", "Description", "Disabled", "Locked"]
USER_TABLE_FIELDS = ["name", "full_name", "description", "disabled", "locked"]


def _row_generator(data):
    return [data["name"], data["full_name"], data["description"], str(not data["disabled"]), str(data["locked"])]


def _parse_users(json_output: List[Dict[str, Any]]) -> List[User]:
    return [UserData(**user).__dict__ for user in json_output]


@winrm
@member_only
def get_all(windows_server: WindowsServer) -> Union[List[User], list]:
    """Get all local users from a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    A list of [User](../inputs_outputs/#user) dictionaries.

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users
    ```
"""
    output = client.run_powershell(
        windows_server, script="Get-WmiObject -Class Win32_UserAccount -Filter \"LocalAccount='True'\" | ConvertTo-JSON"
    )

    if not output:
        message = "Failed to get users as the response from the server was empty."
        ui.output_error(message)
        raise ValueError(f"{message} Output: {repr(output)}.")
    return _parse_users(json_load(output))


@winrm
@member_only
def create(
    windows_server: WindowsServer,
    name: str,
    full_name: str,
    password: str,
    enabled: Optional[bool] = True,
    description: Optional[str] = None,
    password_never_expires: Optional[bool] = False,
) -> None:
    """Create a user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ the username

    - __full_name:__ new user's full name

    - __password:__ a password for the user

    - __enabled:__ set to `false` to create the user in a disabled state (defaults to `true`)

    - __description:__ an optional description for the user

    - __password_never_expires:__ set to `true` to prevent the user's password expiring (defaults to `false`)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.create:
        name: example.user
        full_name: Example User
        description: Create by PPA
      load:
        password: new_password
        windows_server: windows_server
    ```
"""
    command_chunks = [
        f"New-LocalUser -Name '{name}' -FullName '{full_name}' -Password (ConvertTo-SecureString -AsPlainText -Force '{password}')"
    ]
    if enabled is False:
        command_chunks.append("-Disabled")
    if description is not None:
        command_chunks.append(f"-Description '{description}'")
    if password_never_expires:
        command_chunks.append("-PasswordNeverExpires")
    with error_handler(
        errors=[
            {
                "original": "already exists",
                "new": f"Failed to create the user as another with name '{name}' already exists.",
                "exception": exceptions.AlreadyExists,
            },
            {
                "original": "new password does not meet the length",
                "new": "Failed to set the new password as it does not meet complexity requirements.",
                "exception": exceptions.InvalidPassword,
            },
        ]
    ):
        client.run_powershell(windows_server, script=" ".join(command_chunks))


@winrm
@member_only
def set_password(windows_server: WindowsServer, name: str, password: str) -> None:
    """Set a password for a user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ username of the user to enable

    - __password:__ a new password to set

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.set_password:
        name: example.user
      load:
        password: new_password
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "new password does not meet the length",
                "new": "Failed to set the new password as it does not meet complexity requirements.",
                "exception": exceptions.InvalidPassword,
            },
            {
                "original": "was not found",
                "new": f"Failed to set new password as user '{name}' does not exist.",
                "exception": exceptions.DoesNotExist,
            },
        ]
    ):
        client.run_powershell(
            windows_server,
            script=f"Set-LocalUser -Name '{name}' -Password (ConvertTo-SecureString -AsPlainText -Force '{password}')",
        )


@winrm
@member_only
def add_to_group(windows_server: WindowsServer, user_name: str, group_name: str) -> None:
    r"""Add a local or domain user to a group on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __user_name:__ the name of a local user, or _domain\username_ for a domain user

    - __group_name:__ the name of the group to add them to

??? success "Output"

    Nothing is outputted by this action.

??? example "Adding a Local User"

    ```YAML linenums="1"
    - windows_server.users.add_to_group:
        user_name: example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```

??? example "Adding a Domain User"

    ```YAML linenums="1"
    - windows_server.users.add_to_group:
        user_name: domain\example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {group_name} was not found",
                "new": f"Failed to add '{user_name}' to '{group_name}' as the group does not exist.",
                "exception": exceptions.DoesNotExist,
            },
            {
                "original": f"Principal {user_name} was not found",
                "new": f"Failed to add '{user_name}' to '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
            },
        ],
        ignore=["already a member of group"],
    ):
        client.run_powershell(windows_server, script=f'Add-LocalGroupMember -Group "{group_name}" -Member "{user_name}"')


@winrm
@member_only
def remove_from_group(windows_server: WindowsServer, user_name: str, group_name: str) -> None:
    r"""Remove a local or domain user from a group on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __user_name:__ the name of a local user, or _domain\username_ for a domain user

    - __group_name:__ the name of the group to remove them from

??? success "Output"

    Nothing is outputted by this action.

??? example "Removing a Local User"

    ```YAML linenums="1"
    - windows_server.users.remove_from_group:
        user_name: example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```

??? example "Removing a Domain User"

    ```YAML linenums="1"
    - windows_server.users.remove_from_group:
        user_name: domain\example.user
        group_name: example.group
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": f"Group {group_name} was not found",
                "new": f"Failed to remove '{user_name}' from '{group_name}' as the group does not exist.",
                "exception": exceptions.DoesNotExist,
            },
            {
                "original": f"Principal {user_name} was not found",
                "new": f"Failed to remove '{user_name}' from '{group_name}' as the user does not exist.",
                "exception": exceptions.DoesNotExist,
            },
        ],
        ignore=["was not found in group"],
    ):
        client.run_powershell(windows_server, script=f'Remove-LocalGroupMember -Group "{group_name}" -Member "{user_name}"')


@winrm
@member_only
def delete(windows_server: WindowsServer, name: str) -> None:
    """Delete a user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ username of the user to delete

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.delete:
        name: example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"Failed to delete user '{name}' as it does not exist.",
                "exception": exceptions.DoesNotExist,
            }
        ]
    ):
        client.run_powershell(windows_server, script=f'Remove-LocalUser -Name "{name}"')


@winrm
@member_only
def enable(windows_server: WindowsServer, name: str) -> None:
    """Enable a user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ username of the user to enable

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.enable:
        name: example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"Failed to enable user '{name}' as it does not exist.",
                "exception": exceptions.DoesNotExist,
            }
        ]
    ):
        client.run_powershell(windows_server, script=f'Enable-LocalUser -Name "{name}"')


@winrm
@member_only
def disable(windows_server: WindowsServer, name: str) -> None:
    """Disable a user on a Windows Server.

!!! question "Minimum Plugin Version: 7.0.0"

!!! danger "Domain Controllers"

    This action will fail if run against a Domain Controller.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ username of the user to disable

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.users.disable:
        name: example.user
      load:
        windows_server: windows_server
    ```
"""
    with error_handler(
        errors=[
            {
                "original": "was not found",
                "new": f"Failed to disable user '{name}' as it does not exist.",
                "exception": exceptions.DoesNotExist,
            }
        ]
    ):
        client.run_powershell(windows_server, script=f'Disable-LocalUser -Name "{name}"')


def display(text: str, users: Union[User, List[User]]) -> None:
    """Display users in a table.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - windows_server.users.display:
        text: All Local Users
      load:
        users: all_users
    ```
    """
    tables.output_table(text, header=USER_TABLE_HEADER, extract=_row_generator, records=users)


def display_tabbed(text: str, tabs: Dict[str, Union[User, List[User]]]) -> None:
    """Display users in a table with multiple tabs.

This action can be used to display user lists from multiple servers.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_users
        dictionary: tabs
      save: tabs

    - windows_server.users.display_tabbed:
        text: All Local Users
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(
        text, header=USER_TABLE_HEADER, extract=_row_generator, tabs=tabs,
    )


def select(
    text: str, users: Union[User, List[User]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display users in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Users](../inputs_outputs/#user)

    - **all**: a list of selected [Users](../inputs_outputs/#user)

    - **first**: the first selected [User](../inputs_outputs/#user)

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: users

    - windows_server.users.select:
        text: Select Users
      load:
        users: users
      save: selection
    ```
    """
    return tables.input_table(
        text, header=USER_TABLE_HEADER, extract=_row_generator, records=users, minimum=minimum, maximum=maximum
    )


def select_one(text: str, users: Union[User, List[User]]) -> User:
    """Display users in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    A single [User](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: users

    - windows_server.users.select_one:
        text: Select a User
      load:
        users: users
      save: user
    ```
    """
    return select(text, users, minimum=1, maximum=1)["first"]
